#include <stdio.h>
#include <math.h>
#include "mkl.h"

//   
#define BIG_SEED      2000000000
//   ( )
#define BASE_GEN      VSL_BRNG_MCG59
//     
#define GAUSS_METHOD  VSL_METHOD_DGAUSSIAN_ICDF

//  
#define NPATHS    1000

//  
#define NSTEPS    512

//  (%  0.2 -> 20%)
#define SIG   0.2
//   (%  0.05 -> -5%)
#define R     0.05
//    (3 -> 3  )
#define TIME  3.0
//      t=0 (..)
#define S0    100.0

//  
VSLStreamStatePtr InitGen()
{
  VSLStreamStatePtr stream;
  const unsigned int seed[2] = {BIG_SEED, BIG_SEED};
  //  
  vslNewStreamEx(&stream, BASE_GEN, 2, seed);
  return stream;
}

//    
void FreeGen(VSLStreamStatePtr stream)
{
  vslDeleteStream(&stream);
}

//    N(a, b)
//   z  count    
//  ,   
//      
// stream -       MKL
void GenerateGauss(double a, double b, int count, VSLStreamStatePtr stream, double *z)
{
  // o count        z
  vdRngGaussian(GAUSS_METHOD, stream, count, z, a, b);
}

//    csv-
void PrintVectorToFile(double *buffer, int n, char * FileName)
{
  FILE *f = fopen(FileName, "w");
  for (int i =0; i < n; i++)
  {
    fprintf(f, "%lf;\n", buffer[i]);
  }
  fclose(f);
}

//    csv-,     
void PrintVectorToFileEx(double *buffer, int n, char * FileName)
{
  FILE *f = fopen(FileName, "w");
  for (int i =0; i < n; i+=2)
  {
    fprintf(f, "%lf;%lf;\n", buffer[i], buffer[i+1]);
  }
  fclose(f);
}

//    csv-
void PrintMatrixToFile(double **buffer, int nRows, int nColumns, char * FileName)
{
  FILE *f = fopen(FileName, "w");
  for (int i =0; i < nRows; i++)
  {
    for (int j = 0; j < nColumns; j++)
      fprintf(f, "%lf;", buffer[i][j]);
    fprintf(f, "\n");
  }
  fclose(f);
}

//    
// dS(t) = S(t) (r(t) dt + sig(t) dWt), S(0) > 0 ( S(0) = S0)

// S(t) -  
// W -   :
//    W(0) = 0
//    W(t) -    1
//    W(t) - W(s) -  N(0, t-s),  0 <= s < t
// r -  
// sig - 

//  r(t) = const, sig(t) = const   :
// S(Time) = S(0) * exp ((r - sig * sig / 2.0) * Time + sig * z),
//  z -   , z  N(0, Time)

//    S(t)
// z = Wt
double GetStockPrice(double z, double t)
{
  return S0 * exp( (R - SIG * SIG / 2.0) * t + SIG * z);
}


//    
// stream -     
// nSteps -      [0; Time]
// buffer -  W(t)  nSteps+1  ,  
void SimulateWienerProcess(VSLStreamStatePtr stream, int nSteps, double Time, double *buffer)
{
  //   
  double *dw = new double[nSteps];
  //   
  double h = Time / (double)nSteps;
  //  ,   nSteps  
  GenerateGauss(0, sqrt(h), nSteps, stream, dw);
  buffer[0] = 0;
  //    
  for (int j = 1; j <= nSteps; j++)
  {
    buffer[j] = buffer[j - 1] + dw[j - 1];
  }
  //  
  delete [] dw;
}

//       Zt
// stream -     
// nSteps -      [0; Time]
// buffer -  W(t)  Z(t)  nSteps+1  ,  
void SimulateWandZProcesses(VSLStreamStatePtr stream, int nSteps, double Time, double *buffer)
{
  //   
  double *dw = new double[nSteps * 2];
  //   
  double h = Time / (double)nSteps;

  //  ,   nSteps    2
  double mean[2]; mean[0] = 0.0; mean[1] = 0.0;
  double hh = h * sqrt(h);
  double cov[3]; cov[0] = sqrt(h); cov[1] = 0.5 * hh;  
         cov[2] = 1 / sqrt(12.0) * hh;
  vdRngGaussianMV(GAUSS_METHOD, stream, nSteps, dw, 2, 
                  VSL_MATRIX_STORAGE_PACKED, mean, cov);
  //PrintVectorToFileEx(dw, nSteps * 2, "test.csv");
  
  buffer[0] = 0; buffer[1] = 0;
  //    
  for (int j = 2; j <= nSteps * 2; j+=2)
  {
    buffer[j] = buffer[j - 2] + dw[j - 2];
    buffer[j+1] = buffer[j - 1] + dw[j - 1];
  }
  //  
  delete [] dw;
}

//   Euler-Maruyama
double DoEMStep(double S, double dt, double dw)
{
  return S + S * (R * dt + SIG * dw);
}

//   Milstein
double DoMStep(double S, double dt, double dw)
{
  return S + S * (R * dt + SIG * dw) + 0.5 * S * SIG * SIG * (dw * dw - dt);
}


//   S(t)   
void SimulateStockPricesNumAn(VSLStreamStatePtr stream, int nPaths, int nSteps, double Time, double *Error)
{
  //      
  double *wBuffer = new double[nSteps + 1];

  //   
  for (int i = 0; i < nPaths; i++)
  {
    //    
    //     W(t)
    //       
    SimulateWienerProcess(stream, nSteps, Time, wBuffer);
    //   S(Time)   
    double S_An = GetStockPrice(wBuffer[nSteps], Time);
    //   S(Time) ,   
    // nSteps/128, nSteps/64, nSteps/32, nSteps/16, nSteps/8, 
    // nSteps/4,   nSteps/2,  nSteps  
    double S_Num[8];
    int scale = 128;
    for (int j = 0; j < 8; j++)
    {
      S_Num[j] = S0;
      int numMethodSteps = nSteps / scale; //  ASSERT!!!
      double h = Time / (double)numMethodSteps;
      double t = 0;
      int index = 0;
      for (int k = 0; k < numMethodSteps; k++)
      {
        t = t + h;
        index = index + scale;
        //  DoMStep  DoEMStep   
        S_Num[j] = DoMStep(S_Num[j], h, wBuffer[index] - wBuffer[index - scale]);
      }
      Error[j] = Error[j] + fabs(S_An - S_Num[j]);
      scale = scale / 2;
    }
  }
  //  
  for (int j = 0; j < 8; j++)
    Error[j] = Error[j] / nPaths;

  delete [] wBuffer;
}

//   S(t)   
//    -  1.5
void SimulateStockPricesNumAnRK(VSLStreamStatePtr stream, int nPaths, int nSteps, double Time, double *Error)
{
  //      
  //    Zt
  double *wBuffer = new double[(nSteps + 1) * 2];

  //   
  for (int i = 0; i < nPaths; i++)
  {
    //    
    //    Zt
    //     W(t)  Z(t)
    //       
    SimulateWandZProcesses(stream, nSteps, Time, wBuffer);
    //   S(Time)   
    double S_An = GetStockPrice(wBuffer[nSteps * 2], Time);
    //   S(Time) ,   
    // nSteps/128, nSteps/64, nSteps/32, nSteps/16, nSteps/8, 
    // nSteps/4,   nSteps/2,  nSteps  
    double S_Num[8];
    int scale = 128;
    for (int j = 0; j < 8; j++)
    {
      S_Num[j] = S0;
      int numMethodSteps = nSteps / scale; //  ASSERT!!!
      double h = Time / (double)numMethodSteps;
      double t = 0;
      int index = 0;
      for (int k = 0; k < numMethodSteps; k++)
      {
        t = t + h;
        index = index + scale;
        double dw = wBuffer[index * 2] - wBuffer[(index - scale) * 2];
        double dz = wBuffer[index * 2 + 1] - wBuffer[(index - scale) * 2 + 1];

        //   -
        S_Num[j] = S_Num[j] * (1.0 + R * h + SIG * dw) + 
                   0.5 * S_Num[j] * SIG * SIG * (dw * dw - h) +
                   R * SIG * S_Num[j] * dz +
                   0.5 * R * R * S_Num[j] * h * h +
                   R * SIG * S_Num[j] * (dw * h - dz) +
                   0.5 * SIG * SIG * SIG * S_Num[j] * (1.0 / 3.0 * dw * dw - h) * dw;
      }
      Error[j] = Error[j] + fabs(S_An - S_Num[j]);
      scale = scale / 2;
    }
  }
  //  
  for (int j = 0; j < 8; j++)
    Error[j] = Error[j] / nPaths;

  delete [] wBuffer;
}


int main()
{
  //  
  VSLStreamStatePtr stream = InitGen();

  double *Error = new double[8];
  for (int i = 0; i < 8; i++)
    Error[i] = 0.0;

  SimulateStockPricesNumAn(stream, NPATHS, NSTEPS, TIME, Error);
  for (int i = 0; i < 8; i++)
    printf("Error %d = %lf\n", i + 1, Error[i]);

  delete [] Error;

  FreeGen(stream);

  return 0;
}
